/****************************************************************************
** Meta object code from reading C++ file 'game.h'
**
** Created by: The Qt Meta Object Compiler version 67 (Qt 5.9.4)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../src/kubrick-17.12.3/src/game.h"
#include <QtCore/qbytearray.h>
#include <QtCore/qmetatype.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'game.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 67
#error "This file was generated using the moc from 5.9.4. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

QT_BEGIN_MOC_NAMESPACE
QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
struct qt_meta_stringdata_Game_t {
    QByteArrayData data[38];
    char stringdata0[397];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_Game_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_Game_t qt_meta_stringdata_Game = {
    {
QT_MOC_LITERAL(0, 0, 4), // "Game"
QT_MOC_LITERAL(1, 5, 9), // "newPuzzle"
QT_MOC_LITERAL(2, 15, 0), // ""
QT_MOC_LITERAL(3, 16, 7), // "undoAll"
QT_MOC_LITERAL(4, 24, 4), // "load"
QT_MOC_LITERAL(5, 29, 4), // "save"
QT_MOC_LITERAL(6, 34, 6), // "saveAs"
QT_MOC_LITERAL(7, 41, 12), // "changePuzzle"
QT_MOC_LITERAL(8, 54, 19), // "Kubrick::PuzzleItem"
QT_MOC_LITERAL(9, 74, 6), // "puzzle"
QT_MOC_LITERAL(10, 81, 13), // "newCubeDialog"
QT_MOC_LITERAL(11, 95, 8), // "undoMove"
QT_MOC_LITERAL(12, 104, 8), // "redoMove"
QT_MOC_LITERAL(13, 113, 10), // "toggleDemo"
QT_MOC_LITERAL(14, 124, 9), // "solveCube"
QT_MOC_LITERAL(15, 134, 15), // "setStandardView"
QT_MOC_LITERAL(16, 150, 7), // "redoAll"
QT_MOC_LITERAL(17, 158, 11), // "changeScene"
QT_MOC_LITERAL(18, 170, 12), // "cycleSceneUp"
QT_MOC_LITERAL(19, 183, 14), // "cycleSceneDown"
QT_MOC_LITERAL(20, 198, 14), // "watchShuffling"
QT_MOC_LITERAL(21, 213, 10), // "watchMoves"
QT_MOC_LITERAL(22, 224, 14), // "enableMessages"
QT_MOC_LITERAL(23, 239, 13), // "optionsDialog"
QT_MOC_LITERAL(24, 253, 16), // "switchBackground"
QT_MOC_LITERAL(25, 270, 11), // "setMoveAxis"
QT_MOC_LITERAL(26, 282, 1), // "i"
QT_MOC_LITERAL(27, 284, 12), // "setMoveSlice"
QT_MOC_LITERAL(28, 297, 5), // "slice"
QT_MOC_LITERAL(29, 303, 16), // "setMoveDirection"
QT_MOC_LITERAL(30, 320, 9), // "direction"
QT_MOC_LITERAL(31, 330, 7), // "smInput"
QT_MOC_LITERAL(32, 338, 6), // "smCode"
QT_MOC_LITERAL(33, 345, 7), // "advance"
QT_MOC_LITERAL(34, 353, 14), // "addPlayersMove"
QT_MOC_LITERAL(35, 368, 5), // "Move*"
QT_MOC_LITERAL(36, 374, 4), // "move"
QT_MOC_LITERAL(37, 379, 17) // "setCubeNotAligned"

    },
    "Game\0newPuzzle\0\0undoAll\0load\0save\0"
    "saveAs\0changePuzzle\0Kubrick::PuzzleItem\0"
    "puzzle\0newCubeDialog\0undoMove\0redoMove\0"
    "toggleDemo\0solveCube\0setStandardView\0"
    "redoAll\0changeScene\0cycleSceneUp\0"
    "cycleSceneDown\0watchShuffling\0watchMoves\0"
    "enableMessages\0optionsDialog\0"
    "switchBackground\0setMoveAxis\0i\0"
    "setMoveSlice\0slice\0setMoveDirection\0"
    "direction\0smInput\0smCode\0advance\0"
    "addPlayersMove\0Move*\0move\0setCubeNotAligned"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_Game[] = {

 // content:
       7,       // revision
       0,       // classname
       0,    0, // classinfo
      28,   14, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

 // slots: name, argc, parameters, tag, flags
       1,    0,  154,    2, 0x0a /* Public */,
       3,    0,  155,    2, 0x0a /* Public */,
       4,    0,  156,    2, 0x0a /* Public */,
       5,    0,  157,    2, 0x0a /* Public */,
       6,    0,  158,    2, 0x0a /* Public */,
       7,    1,  159,    2, 0x0a /* Public */,
      10,    0,  162,    2, 0x0a /* Public */,
      11,    0,  163,    2, 0x0a /* Public */,
      12,    0,  164,    2, 0x0a /* Public */,
      13,    0,  165,    2, 0x0a /* Public */,
      14,    0,  166,    2, 0x0a /* Public */,
      15,    0,  167,    2, 0x0a /* Public */,
      16,    0,  168,    2, 0x0a /* Public */,
      17,    1,  169,    2, 0x0a /* Public */,
      18,    0,  172,    2, 0x0a /* Public */,
      19,    0,  173,    2, 0x0a /* Public */,
      20,    0,  174,    2, 0x0a /* Public */,
      21,    0,  175,    2, 0x0a /* Public */,
      22,    0,  176,    2, 0x0a /* Public */,
      23,    0,  177,    2, 0x0a /* Public */,
      24,    0,  178,    2, 0x0a /* Public */,
      25,    1,  179,    2, 0x0a /* Public */,
      27,    1,  182,    2, 0x0a /* Public */,
      29,    1,  185,    2, 0x0a /* Public */,
      31,    1,  188,    2, 0x0a /* Public */,
      33,    0,  191,    2, 0x08 /* Private */,
      34,    1,  192,    2, 0x08 /* Private */,
      37,    0,  195,    2, 0x08 /* Private */,

 // slots: parameters
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, 0x80000000 | 8,    9,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, QMetaType::Int,    2,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, QMetaType::Int,   26,
    QMetaType::Void, QMetaType::Int,   28,
    QMetaType::Void, QMetaType::Int,   30,
    QMetaType::Void, QMetaType::Int,   32,
    QMetaType::Void,
    QMetaType::Void, 0x80000000 | 35,   36,
    QMetaType::Void,

       0        // eod
};

void Game::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        Game *_t = static_cast<Game *>(_o);
        Q_UNUSED(_t)
        switch (_id) {
        case 0: _t->newPuzzle(); break;
        case 1: _t->undoAll(); break;
        case 2: _t->load(); break;
        case 3: _t->save(); break;
        case 4: _t->saveAs(); break;
        case 5: _t->changePuzzle((*reinterpret_cast< const Kubrick::PuzzleItem(*)>(_a[1]))); break;
        case 6: _t->newCubeDialog(); break;
        case 7: _t->undoMove(); break;
        case 8: _t->redoMove(); break;
        case 9: _t->toggleDemo(); break;
        case 10: _t->solveCube(); break;
        case 11: _t->setStandardView(); break;
        case 12: _t->redoAll(); break;
        case 13: _t->changeScene((*reinterpret_cast< const int(*)>(_a[1]))); break;
        case 14: _t->cycleSceneUp(); break;
        case 15: _t->cycleSceneDown(); break;
        case 16: _t->watchShuffling(); break;
        case 17: _t->watchMoves(); break;
        case 18: _t->enableMessages(); break;
        case 19: _t->optionsDialog(); break;
        case 20: _t->switchBackground(); break;
        case 21: _t->setMoveAxis((*reinterpret_cast< int(*)>(_a[1]))); break;
        case 22: _t->setMoveSlice((*reinterpret_cast< int(*)>(_a[1]))); break;
        case 23: _t->setMoveDirection((*reinterpret_cast< int(*)>(_a[1]))); break;
        case 24: _t->smInput((*reinterpret_cast< const int(*)>(_a[1]))); break;
        case 25: _t->advance(); break;
        case 26: _t->addPlayersMove((*reinterpret_cast< Move*(*)>(_a[1]))); break;
        case 27: _t->setCubeNotAligned(); break;
        default: ;
        }
    }
}

const QMetaObject Game::staticMetaObject = {
    { &QObject::staticMetaObject, qt_meta_stringdata_Game.data,
      qt_meta_data_Game,  qt_static_metacall, nullptr, nullptr}
};


const QMetaObject *Game::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Game::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_Game.stringdata0))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int Game::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 28)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 28;
    } else if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 28)
            *reinterpret_cast<int*>(_a[0]) = -1;
        _id -= 28;
    }
    return _id;
}
QT_WARNING_POP
QT_END_MOC_NAMESPACE
